from policyengine_uk.model_api import *


class sda(Variable):
    value_type = float
    entity = Person
    label = "Severe Disablement Allowance"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        reported = person("SDA_reported", period) > 0
        # SDA recipients receive a basic rate, and potentially
        # an age-related addition. We assume they receive the highest
        # age-related addition.
        rate = parameters(period).gov.dwp.sda.maximum
        return reported * rate * WEEKS_IN_YEAR


class SDA_reported(Variable):
    value_type = float
    entity = Person
    label = "Severe Disablement Allowance (reported)"
    definition_period = YEAR
    unit = GBP



from policyengine_uk.model_api import *

label = "Region"


class Region(Enum):
    UNKNOWN = "Unknown"
    NORTH_EAST = "North East"
    NORTH_WEST = "North West"
    YORKSHIRE = "Yorkshire and the Humber"
    EAST_MIDLANDS = "East Midlands"
    WEST_MIDLANDS = "West Midlands"
    EAST_OF_ENGLAND = "East of England"
    LONDON = "London"
    SOUTH_EAST = "South East"
    SOUTH_WEST = "South West"
    WALES = "Wales"
    SCOTLAND = "Scotland"
    NORTHERN_IRELAND = "Northern Ireland"



from policyengine_uk.model_api import *

label = "Maintenance"


class housing_service_charges(Variable):
    value_type = float
    entity = Household
    label = "housing service charges"
    documentation = "Total amount spent on housing service charges"
    definition_period = YEAR
    unit = GBP


class water_and_sewerage_charges(Variable):
    value_type = float
    entity = Household
    label = "water and sewerage charges"
    documentation = "Total amount spent on water and sewerage charges"
    definition_period = YEAR


class employer_pension_contributions(Variable):
    value_type = float
    entity = Person
    label = "Employer pension contributions"
    documentation = "Total amount spent on employer pension contributions"
    definition_period = YEAR
    unit = GBP


class maintenance_expenses(Variable):
    value_type = float
    entity = Person
    label = "maintenance expenses"
    definition_period = YEAR
    unit = GBP


class mortgage_interest_repayment(Variable):
    value_type = float
    entity = Household
    label = "mortgage interest repayments"
    documentation = "Total amount spent on mortgage interest repayments"
    definition_period = YEAR
    unit = GBP


class mortgage_capital_repayment(Variable):
    value_type = float
    entity = Household
    label = "mortgage capital repayments"
    definition_period = YEAR
    unit = GBP


class council_tax(Variable):
    value_type = float
    entity = Household
    label = "Council Tax"
    documentation: str = "Gross amount spent on Council Tax, before discounts"
    definition_period = YEAR
    unit = GBP
    quantity_type = FLOW
    uprating: str = "calibration.uprating.council_tax"



from policyengine_uk.model_api import *


class is_carer_for_benefits(Variable):
    value_type = bool
    entity = Person
    label = "Whether this person is a carer for benefits purposes"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("receives_carers_allowance", period)


class benunit_has_carer(Variable):
    value_type = bool
    entity = BenUnit
    label = "Benefit unit has a carer"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit("num_carers", period) > 0


class num_carers(Variable):
    value_type = int
    entity = BenUnit
    label = "Number of carers in the family"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return aggr(benunit, period, ["is_carer_for_benefits"])


class carer_premium(Variable):
    value_type = float
    entity = BenUnit
    label = "Carer premium"
    definition_period = YEAR
    reference = (
        "The Social Security Amendment (Carer Premium) Regulations 2002"
    )
    unit = GBP

    def formula(benunit, period, parameters):
        carers = benunit("num_carers", period.this_year)
        CP = parameters(period).gov.dwp.carer_premium
        weekly_premium = select(
            [carers == 0, carers == 1, carers == 2],
            [0, CP.single, CP.couple],
        )
        return weekly_premium * WEEKS_IN_YEAR



from policyengine_uk.model_api import *


class household_owns_tv(Variable):
    label = "Owns a TV"
    documentation = "Whether this household owns a functioning colour TV."
    entity = Household
    definition_period = YEAR
    value_type = bool

    def formula(household, period, parameters):
        percent_owning_tv = parameters(
            period
        ).gov.dcms.bbc.tv_licence.tv_ownership
        return random(household) <= percent_owning_tv



from policyengine_uk.model_api import *


class maternity_allowance_reported(Variable):
    value_type = float
    entity = Person
    label = "Maternity allowance"
    definition_period = YEAR
    unit = GBP


class maternity_allowance(Variable):
    label = "Maternity Allowance"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(person, period, parameters):
        return person("maternity_allowance_reported", period)


class ssmg_reported(Variable):
    label = "Sure Start Maternity Grant (reported)"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP


class ssmg(Variable):
    label = "Sure Start Maternity Grant"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(person, period, parameters):
        return person("ssmg_reported", period)



from policyengine_uk.model_api import *


class savings_credit_income(Variable):
    label = "Income for Savings Credit"
    documentation = "Savings Credit (Pension Credit) excludes certain income sources from the calculation of the amount."
    entity = BenUnit
    definition_period = YEAR
    value_type = float
    unit = GBP
    reference = "https://www.legislation.gov.uk/uksi/2002/1792/regulation/9"

    def formula(benunit, period, parameters):
        guarantee_credit_income = benunit("pension_credit_income", period)
        pc = parameters(period).gov.dwp.pension_credit
        excluded_income = add(
            benunit, period, pc.savings_credit.excluded_income
        )
        return max_(0, guarantee_credit_income - excluded_income)



from policyengine_uk.model_api import *


class state_id(Variable):
    label = "State ID"
    documentation = "Identity of the state"
    entity = State
    definition_period = YEAR
    value_type = int


class person_state_id(Variable):
    label = "State ID"
    documentation = "Identity of the state"
    entity = Person
    definition_period = YEAR
    value_type = int


class person_state_role(Variable):
    label = "State role"
    entity = Person
    definition_period = YEAR
    value_type = str


class state_weight(Variable):
    label = "State weight"
    documentation = "Weight value"
    entity = State
    definition_period = YEAR
    value_type = float



from policyengine_uk.model_api import *


class pension_credit_reported(Variable):
    label = "Pension Credit (reported)"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP



from policyengine_uk.model_api import *


class land_value(Variable):
    label = "Land value"
    documentation = (
        "Estimated total land value (directly and indirectly owned)"
    )
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP
    quantity_type = STOCK

    def formula(household, period):
        VARIABLES = [
            "household_land_value",
            "corporate_land_value",
        ]
        return add(household, period, VARIABLES)


class household_land_value(Variable):
    entity = Household
    label = "Land value"
    documentation = (
        "Estimated total land value directly owned by the household"
    )
    unit = GBP
    definition_period = YEAR
    value_type = float
    quantity_type = STOCK

    def formula(household, period, parameters):
        wealth = parameters(period).household.wealth
        property_wealth_intensity = (
            wealth.land.value.aggregate_household_land_value
            / wealth.property_wealth
        )
        property_wealth = household("property_wealth", period)
        owned_land = household("owned_land", period)
        return property_wealth * property_wealth_intensity + owned_land


class corporate_land_value(Variable):
    entity = Household
    label = "Land value"
    documentation = "Estimated total land value indirectly owned by the household from corporate holdings"
    unit = GBP
    definition_period = YEAR
    value_type = float
    quantity_type = STOCK

    def formula(household, period, parameters):
        wealth = parameters(period).household.wealth
        corporate_wealth = household("corporate_wealth", period)
        corporate_wealth_intensity = (
            wealth.land.value.aggregate_corporate_land_value
            / wealth.corporate_wealth
        )
        return corporate_wealth * corporate_wealth_intensity



from typing import Dict
from policyengine_uk.data.datasets.frs.frs import FRS
from numpy.typing import ArrayLike
from policyengine_core.data.dataset import Dataset

LEGACY_BENEFITS = [
    "working_tax_credit",
    "child_tax_credit",
    "housing_benefit",
    "income_support",
    "ESA_income",
    "JSA_income",
]


def migrate_to_universal_credit(
    dataset: Dataset = FRS, year: int = 2022
) -> Dict[str, ArrayLike]:
    """Converts legacy benefit claimants to Universal Credit claimants by switching
    reported amounts.

    Args:
        dataset (type, optional): The dataset to use. Defaults to FRS.
        year (int, optional): The year to use. Defaults to 2022.

    Returns:
        Dict[str, ArrayLike]: Variables with replaced values.
    """
    from policyengine_uk import Microsimulation

    frs = Microsimulation(
        dataset=dataset,
        year=year,
    )

    changes = {
        f"universal_credit_reported/{year}": frs.calc(
            "universal_credit_reported", period=year
        ).values
    }

    for benefit in LEGACY_BENEFITS:
        reported_amount = frs.calc(benefit + "_reported", period=year).values
        changes[f"{benefit}_reported/{year}"] = reported_amount * 0
        changes[f"universal_credit_reported/{year}"] += reported_amount

    return changes



from policyengine_uk.model_api import *


class baseline_pension_credit_entitlement(Variable):
    label = "PC entitlement (baseline)"
    entity = BenUnit
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"



from policyengine_uk.model_api import *


class domestic_rates(Variable):
    label = "domestic rates"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        rates = parameters(period).gov.local_authorities.domestic_rates.rates
        local_authority = household("local_authority", period)
        rate_defined = pd.Series(local_authority.decode_to_str()).isin(
            rates._children
        )
        percent = np.zeros(household.count, dtype=float)
        if any(rate_defined):
            percent[rate_defined] = rates[local_authority[rate_defined]]
            main_residence_value = household("main_residence_value", period)
            return percent * main_residence_value
        else:
            return 0



from policyengine_uk.model_api import *


class tv_licence(Variable):
    label = "TV licence"
    documentation = "Net cost of a TV licence"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP
    reference = "https://www.legislation.gov.uk/ukpga/2003/21/section/363"
    category = TAX

    def formula(household, period, parameters):
        owns_tv = household("household_owns_tv", period)
        discount = household("tv_licence_discount", period)
        would_evade = household("would_evade_tv_licence_fee", period)
        fee = parameters(period).gov.dcms.bbc.tv_licence.colour
        return (owns_tv & ~would_evade) * fee * (1 - discount)



from policyengine_uk.model_api import *


class energy_price_cap_subsidy(Variable):
    label = "Energy price cap subsidy"
    documentation = "Reduction in energy bills due to offsetting the price cap and compensating energy firms."
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        energy_consumption = household("domestic_energy_consumption", period)
        # For each of the four quarters in the next year, calculate the
        # relative change to the price cap against the baseline price cap,
        # and multiply by quarterly energy consumption.
        total_subsidy = 0
        q1_baseline_energy_price_cap = 0
        for quarter in range(1, 5):
            current_quarter = f"{period.start.year}_q{quarter}"
            price_cap = parameters(period).gov.ofgem.price_cap[current_quarter]
            baseline_price_cap = parameters(
                period
            ).baseline.gov.ofgem.price_cap[current_quarter]
            if quarter == 1:
                q1_baseline_energy_price_cap = baseline_price_cap
            relative_change_in_cap = (
                price_cap - baseline_price_cap
            ) / baseline_price_cap
            relative_change_in_energy_consumption = (
                baseline_price_cap / q1_baseline_energy_price_cap
            )
            total_subsidy += (
                -relative_change_in_cap
                * relative_change_in_energy_consumption
                * energy_consumption
                / 4
            )
        return total_subsidy



from policyengine_uk.model_api import *
from policyengine_uk.variables.household.demographic.geography import Region


class household_id(Variable):
    value_type = int
    entity = Household
    label = "ID for the household"
    definition_period = YEAR


class households(Variable):
    value_type = float
    entity = Household
    label = "Variable holding households"
    definition_period = YEAR
    default_value = 1


class num_bedrooms(Variable):
    value_type = int
    entity = Household
    label = "The number of bedrooms in the house"
    definition_period = YEAR


class is_shared_accommodation(Variable):
    value_type = bool
    entity = Household
    label = "Whether the household is shared accommodation"
    definition_period = YEAR


class household_weight(Variable):
    value_type = float
    entity = Household
    label = "Weight factor for the household"
    definition_period = YEAR


class Country(Enum):
    UNKNOWN = "Unknown"
    SCOTLAND = "Scotland"
    ENGLAND = "England"
    WALES = "Wales"
    NORTHERN_IRELAND = "Northern Ireland"


class country(Variable):
    value_type = Enum
    possible_values = Country
    default_value = Country.ENGLAND
    entity = Household
    label = "Country of the UK"
    definition_period = YEAR

    def formula(household, period, parameters):
        region = household("region", period)
        return select(
            [
                region == Region.UNKNOWN,
                region == Region.SCOTLAND,
                region == Region.WALES,
                region == Region.NORTHERN_IRELAND,
                True,
            ],
            [
                Country.UNKNOWN,
                Country.SCOTLAND,
                Country.WALES,
                Country.NORTHERN_IRELAND,
                Country.ENGLAND,
            ],
        )


class TenureType(Enum):
    RENT_FROM_COUNCIL = "Rented from Council"
    RENT_FROM_HA = "Rented from a Housing Association"
    RENT_PRIVATELY = "Rented privately"
    OWNED_OUTRIGHT = "Owned outright"
    OWNED_WITH_MORTGAGE = "Owned with a mortgage"


class ONSTenureType(Enum):
    RENT_FROM_COUNCIL = "Rented from Council"
    RENT_FROM_HA = "Rented from a Housing Association"
    RENT_PRIVATELY = "Rented privately"
    OWNER_OCCUPIED = "Owner occupied"


class tenure_type(Variable):
    value_type = Enum
    possible_values = TenureType
    default_value = TenureType.RENT_PRIVATELY
    entity = Household
    label = "Tenure type of the household"
    definition_period = YEAR


class ons_tenure_type(Variable):
    label = "ONS tenure type"
    documentation = "Tenure type matching ONS_produced statistical breakdowns."
    entity = Household
    definition_period = YEAR
    value_type = Enum
    possible_values = ONSTenureType
    default_value = ONSTenureType.OWNER_OCCUPIED

    def formula(household, period, parameters):
        tenure = household("tenure_type", period)
        return select(
            [
                tenure == TenureType.RENT_FROM_HA,
                tenure == TenureType.RENT_FROM_COUNCIL,
                tenure == TenureType.RENT_PRIVATELY,
                tenure == TenureType.OWNED_OUTRIGHT,
                tenure == TenureType.OWNED_WITH_MORTGAGE,
            ],
            [
                ONSTenureType.RENT_FROM_HA,
                ONSTenureType.RENT_FROM_COUNCIL,
                ONSTenureType.RENT_PRIVATELY,
                ONSTenureType.OWNER_OCCUPIED,
                ONSTenureType.OWNER_OCCUPIED,
            ],
        )


class is_renting(Variable):
    value_type = bool
    entity = Household
    label = "Is renting"
    definition_period = YEAR

    def formula(household, period, parameters):
        tenure = household("tenure_type", period)
        tenures = tenure.possible_values
        RENT_TENURES = [
            tenures.RENT_PRIVATELY,
            tenures.RENT_FROM_COUNCIL,
            tenures.RENT_PRIVATELY,
        ]
        return is_in(tenure, RENT_TENURES)


class AccommodationType(Enum):
    HOUSE_DETACHED = "Detached house"
    HOUSE_SEMI_DETACHED = "Semi-detached house"
    HOUSE_TERRACED = "Terraced house"
    FLAT = "Purpose-built flat or maisonette"
    CONVERTED_HOUSE = "Converted house or building"
    MOBILE = "Caravan/Mobile home or houseboat"
    OTHER = "Other"
    UNKNOWN = "Unknown"


class accommodation_type(Variable):
    value_type = Enum
    possible_values = AccommodationType
    default_value = AccommodationType.UNKNOWN
    entity = Household
    label = "Type of accommodation"
    definition_period = YEAR


class household_equivalisation_bhc(Variable):
    value_type = float
    entity = Household
    label = "Equivalisation factor to account for household composition, before housing costs"
    definition_period = YEAR

    def formula(household, period, parameters):
        count_other_adults = max_(
            household.sum(household.members("is_adult", period)) - 1, 0
        )
        count_young_children = household.sum(
            household.members("is_young_child", period)
        )
        count_older_children = household.sum(
            household.members("is_older_child", period)
        )
        return (
            0.67
            + 0.33 * count_other_adults
            + 0.33 * count_older_children
            + 0.2 * count_young_children
        )


class household_equivalisation_ahc(Variable):
    value_type = float
    entity = Household
    label = "Equivalisation factor to account for household composition, after housing costs"
    definition_period = YEAR

    def formula(household, period, parameters):
        count_other_adults = max_(
            household.sum(household.members("is_adult", period)) - 1, 0
        )
        count_young_children = household.sum(
            household.members("is_young_child", period)
        )
        count_older_children = household.sum(
            household.members("is_older_child", period)
        )
        return (
            0.58
            + 0.42 * count_other_adults
            + 0.42 * count_older_children
            + 0.2 * count_young_children
        )


class household_count_people(Variable):
    value_type = int
    entity = Household
    label = "Number of people"
    definition_period = YEAR
    unit = "person"

    def formula(household, period, parameters):
        return household.nb_persons()


class household_num_benunits(Variable):
    value_type = int
    entity = Household
    label = "Number of benefit units"
    definition_period = YEAR
    unit = "benefit unit"

    def formula(household, period, parameters):
        return household.sum(household.members("is_benunit_head", period))



from policyengine_uk.model_api import *


class LVT(Variable):
    entity = Household
    label = "Land value tax"
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(household, period, parameters):
        lvt = parameters(period).gov.contrib.ubi_center.land_value_tax
        full_lvt = lvt.rate * household("land_value", period)
        household_lvt = lvt.household_rate * household(
            "household_land_value", period
        )
        corporate_lvt = lvt.corporate_rate * household(
            "corporate_land_value", period
        )
        return full_lvt + household_lvt + corporate_lvt



from policyengine_uk.model_api import *

label = "Care"


class receives_carers_allowance(Variable):
    value_type = bool
    entity = Person
    label = "receives Carer's Allowance"
    documentation = "Whether this person receives Carer's Allowance."
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("carers_allowance", period) > 0



from policyengine_uk.model_api import *


class child_minimum_guarantee_addition(Variable):
    label = "Child-related addition"
    entity = BenUnit
    definition_period = YEAR
    value_type = float
    unit = GBP
    reference = "https://www.legislation.gov.uk/uksi/2002/1792/schedule/IIA"

    def formula(benunit, period, parameters):
        person = benunit.members
        is_child = person("is_child_or_QYP", period)
        gc = parameters(period).gov.dwp.pension_credit.guarantee_credit
        standard_disability_benefits = gc.child.disability.eligibility
        severe_disability_benefits = gc.child.disability.severe.eligibility
        is_disabled = add(person, period, standard_disability_benefits) > 0
        is_severely_disabled = (
            add(person, period, severe_disability_benefits) > 0
        )
        is_standard_disabled = is_disabled & ~is_severely_disabled
        is_not_disabled = ~is_disabled
        per_child_amount = (
            select(
                [
                    is_child & is_not_disabled,
                    is_child & is_standard_disabled,
                    is_child & is_severely_disabled,
                ],
                [
                    gc.child.addition,
                    gc.child.addition + gc.child.disability.addition,
                    gc.child.addition + gc.child.disability.severe.addition,
                ],
            )
            * WEEKS_IN_YEAR
        )
        return benunit.sum(per_child_amount)



from policyengine_uk.model_api import *
from policyengine_uk.variables.gov.dwp.pip.pip import PIPCategory


class PIP_DL_reported(Variable):
    value_type = float
    entity = Person
    label = "PIP (self-care) (reported)"
    definition_period = YEAR
    unit = GBP


class pip_dl(Variable):
    label = "PIP (daily living)"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(person, period, parameters):
        pip_dl = parameters(period).gov.dwp.pip.daily_living
        category = person("pip_dl_category", period)
        return (
            select(
                [
                    category == PIPCategory.ENHANCED,
                    category == PIPCategory.STANDARD,
                    category == PIPCategory.NONE,
                ],
                [
                    pip_dl.enhanced,
                    pip_dl.standard,
                    0,
                ],
            )
            * WEEKS_IN_YEAR
        )


class receives_enhanced_pip_dl(Variable):
    label = "Receives enhanced PIP (daily living)"
    entity = Person
    definition_period = YEAR
    value_type = bool

    def formula(person, period, parameters):
        return person("pip_dl_category", period) == PIPCategory.ENHANCED



from policyengine_uk.model_api import *


class BSP(Variable):
    value_type = float
    entity = Person
    label = "Bereavement Support Payment"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        return person("BSP_reported", period)


class BSP_reported(Variable):
    value_type = float
    entity = Person
    label = "Bereavement Support Payment (reported)"
    definition_period = YEAR
    unit = GBP



from policyengine_uk import Microsimulation
import pytest
from policyengine_uk.data import EnhancedFRS

YEARS = range(2022, 2026)

baseline = Microsimulation(dataset=EnhancedFRS, dataset_year=2022)


@pytest.mark.parametrize("year", YEARS)
def test_not_nan(year):
    for variable in baseline.tax_benefit_system.variables:
        assert ~baseline.calc(variable, period=year).isna().any()



from policyengine_uk.model_api import *


class incapacity_benefit(Variable):
    value_type = float
    entity = Person
    label = "Incapacity Benefit"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        return person("incapacity_benefit_reported", period)


class incapacity_benefit_reported(Variable):
    value_type = float
    entity = Person
    label = "Incapacity Benefit (reported)"
    definition_period = YEAR
    unit = GBP



import numpy as np
from ...loss_category import LossCategory
import tensorflow as tf
from typing import Iterable, Tuple


class Households(LossCategory):
    name = "Households"
    category = "Demographics"

    def initialise(self):
        region_ct_band_parameter = (
            self.calibration_parameters.demographics.households.by_region_by_council_tax_band
        )

        uk_wide = self.calibration_parameters.demographics.households.in_total

        self.comparisons = []

        ct_band = self.sim.calc("council_tax_band").values
        region = self.sim.calc("region").values

        # Region - CT band

        total_actuals = 0

        for target_region in region_ct_band_parameter._children:
            for target_ct_band in region_ct_band_parameter._children[
                target_region
            ]._children:
                total_actuals += region_ct_band_parameter._children[
                    target_region
                ]._children[target_ct_band]

        region_ct_band_adjustment = uk_wide / total_actuals

        for target_region in region_ct_band_parameter._children:
            for target_ct_band in region_ct_band_parameter._children[
                target_region
            ]._children:
                actual = region_ct_band_parameter._children[
                    target_region
                ]._children[target_ct_band]
                self.comparisons.append(
                    (
                        f"{target_region}_{target_ct_band}",
                        (ct_band == target_ct_band)
                        & (region == target_region),
                        actual * region_ct_band_adjustment,
                    )
                )

        # Region - tenure type

        tenure_type = self.sim.calc("ons_tenure_type").values

        region_tenure_parameter = (
            self.calibration_parameters.demographics.households.by_region_by_tenure_type
        )
        total_actuals = 0
        for target_region in region_tenure_parameter._children:
            regional_population = 0
            for target_tenure_type in region_tenure_parameter._children[
                target_region
            ]._children:
                total_actuals += region_tenure_parameter._children[
                    target_region
                ]._children[target_tenure_type]

        region_tenure_adjustment = uk_wide / total_actuals

        for target_region in region_tenure_parameter._children:
            regional_population = 0
            for target_tenure_type in region_tenure_parameter._children[
                target_region
            ]._children:
                actual_population = region_tenure_parameter._children[
                    target_region
                ]._children[target_tenure_type]
                self.comparisons.append(
                    (
                        f"{target_region}_{target_tenure_type}",
                        (tenure_type == target_tenure_type)
                        & (region == target_region),
                        actual_population * region_tenure_adjustment,
                    )
                )
                regional_population += actual_population

            self.comparisons += [
                (
                    f"households.{target_region}",
                    region == target_region,
                    regional_population * region_tenure_adjustment,
                )
            ]

        self.comparisons += [
            (
                "households.UNITED_KINGDOM",
                np.ones_like(region),
                uk_wide,
            )
        ]

    def get_loss_subcomponents(
        self, household_weights: tf.Tensor
    ) -> Iterable[Tuple]:
        for name, values, actual in self.comparisons:
            yield (
                name,
                tf.reduce_sum(household_weights * values),
                actual,
            )

    def get_metric_names(self) -> Iterable[str]:
        return [x[0] for x in self.comparisons]



from policyengine_uk.model_api import *


class pension_credit_income(Variable):
    label = "Income for Pension Credit"
    entity = BenUnit
    definition_period = YEAR
    value_type = float
    unit = GBP
    reference = "https://www.legislation.gov.uk/ukpga/2002/16/section/15"

    def formula(benunit, period, parameters):
        pc = parameters(period).gov.dwp.pension_credit
        sources = pc.guarantee_credit.income
        total = add(benunit, period, sources)
        bi = parameters(period).gov.contrib.ubi_center.basic_income
        if bi.interactions.include_in_means_tests:
            total += add(benunit, period, ["basic_income"])
        pension_contributions = add(benunit, period, ["pension_contributions"])
        tax = add(benunit, period, ["income_tax", "national_insurance"])
        pen_con_deduction_rate = pc.income.pension_contributions_deduction
        deductions = tax + pension_contributions * pen_con_deduction_rate
        return max_(0, total - deductions)



from policyengine_uk.model_api import *


class tv_licence_discount(Variable):
    label = "TV licence discount"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "/1"
    reference = (
        "https://www.legislation.gov.uk/ukpga/2003/21/section/365A",
        "https://www.tvlicensing.co.uk/reducedfee",
    )

    def formula(household, period, parameters):
        person = household.members
        tv_licence = parameters(period).gov.dcms.bbc.tv_licence

        # Aged discount
        aged = person("age", period) >= tv_licence.discount.aged.min_age
        has_aged = household.any(aged)
        claims_pc = add(household, period, ["pension_credit"]) > 0
        meets_pc_requirement = (
            ~tv_licence.discount.aged.must_claim_pc | claims_pc
        )
        eligible_for_aged_discount = has_aged & meets_pc_requirement
        aged_discount = (
            eligible_for_aged_discount * tv_licence.discount.aged.discount
        )

        # Blind discount
        is_blind = person("is_blind", period)
        has_blind = household.any(is_blind)
        blind_discount = has_blind * tv_licence.discount.blind.discount

        return max_(aged_discount, blind_discount)



from policyengine_uk.model_api import *


class guarantee_credit(Variable):
    label = "Guarantee Credit"
    entity = BenUnit
    definition_period = YEAR
    value_type = float
    unit = GBP
    reference = "https://www.legislation.gov.uk/ukpga/2002/16/section/2"

    def formula(benunit, period, parameters):
        income = benunit("pension_credit_income", period)
        minimum_guarantee = benunit("minimum_guarantee", period)
        eligible = benunit("is_guarantee_credit_eligible", period)
        return max_(0, minimum_guarantee - income) * eligible



from policyengine_uk.model_api import *


class baseline_business_rates(Variable):
    label = "Baseline business rates incidence"
    documentation = (
        "Total incidence from business rates exposure in the baseline"
    )
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        br = parameters(period).gov.hmrc.business_rates.statistics.revenue
        total_revenue = (
            br.ENGLAND  # HMRC
            + br.SCOTLAND  # Revenue Scotland
            + br.WALES  # Welsh Revenue Authority
            + br.NORTHERN_IRELAND  # HMRC
        )
        return household("shareholding", period) * total_revenue


class business_rates(Variable):
    label = "Business rates incidence"
    documentation = "Total incidence from exposure to business rates via corporate shareholdings"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    formula = baseline_business_rates.formula


class change_in_business_rates(Variable):
    label = "average per-year business rates"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    formula = change_over_baseline(business_rates)


class business_rates_change_incidence(Variable):
    label = "Business rates changes"
    documentation = "Total effect of policy changes to business rates"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period):
        business_rates = household("business_rates", period)
        baseline_business_rates = household("baseline_business_rates", period)
        return business_rates - baseline_business_rates



from policyengine_core.parameters import ParameterNode, Parameter
from policyengine_core.periods import instant


def add_tax_benefit_uprating(parameters: ParameterNode) -> ParameterNode:
    """Adds the uprating index used to uprate most tax-benefit instruments (September CPI of the previous year)."""
    cpi = parameters.calibration.uprating.CPI
    data = {}
    for year in range(2000, 2029):
        data[f"{year + 1}-04-01"] = cpi(instant(f"{year}-09-01"))
    new_parameter = Parameter(
        "uprating.september_cpi",
        data={
            "values": data,
            "metadata": {
                "label": "Tax-benefit CPI",
                "unit": "/1",
                "name": "tax_benefit_cpi",
                "description": "CPI of the previous september (percentage of June 2015 levels). This uprates some (not all) tax-benefit parameters automatically.",
            },
        },
    )
    if not hasattr(parameters.calibration.uprating, "september_cpi"):
        parameters.calibration.uprating.add_child(
            "september_cpi",
            new_parameter,
        )
    else:
        parameters.calibration.uprating.september_cpi = new_parameter

    data = {}
    for year in range(2000, 2029):
        data[f"{year}-01-01"] = cpi(instant(f"{year}-01-01"))
    if not hasattr(parameters.calibration.uprating, "january_cpi"):
        parameters.calibration.uprating.add_child(
            "january_cpi",
            Parameter(
                "uprating.january_cpi",
                data={
                    "values": data,
                },
            ),
        )
    return parameters



import argparse
from pathlib import Path

# This command-line tools enables Plotly charts to show in the HTML files for the Jupyter Book documentation.

parser = argparse.ArgumentParser()
parser.add_argument("book_path", help="Path to the Jupyter Book.")

args = parser.parse_args()

# Find every HTML file in the Jupyter Book. Then, add a script tag to the start of the <head> tag in each file, with the contents:
# <script src="https://cdnjs.cloudflare.com/ajax/libs/require.js/2.3.4/require.min.js"></script>

book_folder = Path(args.book_path)

for html_file in book_folder.glob("**/*.html"):
    with open(html_file, "r") as f:
        html = f.read()

    # Add the script tag to the start of the <head> tag.
    html = html.replace(
        "<head>",
        '<head><script src="https://cdnjs.cloudflare.com/ajax/libs/require.js/2.3.4/require.min.js"></script>',
    )

    with open(html_file, "w") as f:
        f.write(html)



from policyengine_uk.model_api import *


class severe_disability_minimum_guarantee_addition(Variable):
    label = "Severe disability-related increase"
    entity = BenUnit
    definition_period = YEAR
    value_type = float
    unit = GBP
    reference = (
        "https://www.legislation.gov.uk/uksi/2002/1792/schedule/I/paragraph/1"
    )

    def formula(benunit, period, parameters):
        # 1. At least one adult receives a qualifying benefit
        # 2. No children (except children receiving qualifying benefits)
        # 3. Nobody receives Carer's Allowance (technically 'for one of the claimants', but we assume this is true)
        severe_disability = parameters(
            period
        ).gov.dwp.pension_credit.guarantee_credit.severe_disability
        relevant_benefits = severe_disability.relevant_benefits
        person = benunit.members
        person_receives_qualifying_benefits = (
            add(person, period, relevant_benefits) > 0
        )
        is_adult = person("is_adult", period)
        count_eligible_adults = benunit.sum(
            is_adult & person_receives_qualifying_benefits
        )
        any_children_without_benefits = (
            benunit.sum(~is_adult & ~person_receives_qualifying_benefits) > 0
        )
        carers_allowance_received = (
            add(benunit, period, ["carers_allowance"]) > 0
        )
        eligible = ~any_children_without_benefits & ~carers_allowance_received
        return (
            eligible
            * count_eligible_adults
            * severe_disability.addition
            * WEEKS_IN_YEAR
        )



from typing import Iterable
from .households import Households
from .populations import Populations
from ...loss_category import LossCategory


class Demographics(LossCategory):
    name = "Demographics"
    category = "Demographics"
    weight = 1

    def initialise(self):
        self.initial_value = None
        self.subcategories = [
            Households(self.years, self.year, 1, self.sim),
            Populations(self.years, self.year, 1, self.sim),
        ]

    def compute(self, *args, **kwargs):
        losses = 0
        log = []
        for subcategory in self.subcategories:
            additional_loss, additional_log = subcategory.compute(
                *args, **kwargs
            )
            losses += additional_loss * self.weight
            log += additional_log
        if self.initial_value is None:
            self.initial_value = losses
        losses /= self.initial_value
        for entry in log:
            entry["loss"] /= self.initial_value
        return losses, log

    def get_metric_names(self) -> Iterable[str]:
        names = sum(
            [
                subcategory.get_metric_names()
                for subcategory in self.subcategories
            ],
            [],
        )
        dated_names = []
        for name in names:
            for year in self.years:
                dated_names.append(f"{name}.{year}")
        return dated_names



from glob import glob
import os
from time import time
import tensorflow as tf
import numpy as np
from policyengine_uk.data.datasets.frs.enhanced.stages.calibration.calibrated_frs import (
    CalibratedFRS,
)
from policyengine_uk.data.datasets.frs.frs import FRS
from policyengine_uk.data.storage import policyengine_uk_MICRODATA_FOLDER
from policyengine_uk.tools.simulation import Microsimulation
from policyengine_uk.repo import REPO
from policyengine_uk.data.datasets.frs.enhanced.stages.calibration.losses import (
    LossCalculator,
)
from pathlib import Path
import h5py
import pandas as pd
from argparse import ArgumentParser
from policyengine_core.data import Dataset


class HouseholdWeights:
    def __init__(self, start_year: int = 2022, end_year: int = 2027):
        self.start_year = start_year
        self.end_year = end_year

    def calibrate(
        self,
        validation_split: float = 0.1,
        num_epochs: int = 32,
        learning_rate: float = 1e1,
        loss_calculator: LossCalculator = None,
        dataset: Dataset = FRS,
    ):
        """Calibrate FRS weights to administrative statistics.

        Args:
            sim (Microsimulation): A simulation containing demographic variables.
            validation_split (float, optional): The percentage of metrics to use as validation. Defaults to 0.1.
            num_epochs (int, optional): The number of epochs to run. Defaults to 256.
            learning_rate (float, optional): The learning rate for the optimiser. Defaults to 8e+1.
            loss_calculator (LossCalculator, optional): A loss calculator to use. Defaults to None.
            dataset (Dataset, optional): The dataset to use. Defaults to FRS.
        """
        self.sim = (
            loss_calculator.sim
            if loss_calculator is not None
            else Microsimulation(dataset=dataset, average_parameters=True)
        )
        survey_num_households = len(self.sim.calc("household_id"))
        self.weight_changes = tf.Variable(
            np.ones(
                (self.end_year + 1 - self.start_year, survey_num_households)
            ),
            dtype=tf.float32,
        )
        opt = tf.keras.optimizers.Adam(learning_rate=learning_rate)
        loss_calculator = loss_calculator or LossCalculator(
            self.sim,
            validation_split,
            start_year=self.start_year,
            end_year=self.end_year,
        )
        start_train_loss = None
        start_val_loss = None
        start_time = time()
        for epoch in range(num_epochs):
            with tf.GradientTape() as tape:
                loss = loss_calculator.compute_loss(
                    self.weight_changes, validation=False, epoch=epoch
                )
                grads = tape.gradient(loss, self.weight_changes)
            if validation_split > 0:
                validation_loss = loss_calculator.compute_loss(
                    self.weight_changes, validation=True, epoch=epoch
                )
            else:
                validation_loss = tf.constant(0, dtype=tf.float32)
            if start_train_loss is None:
                start_train_loss = loss.numpy()
            if start_val_loss is None:
                start_val_loss = validation_loss.numpy()
            opt.apply_gradients(zip([grads], [self.weight_changes]))
            seconds_elapsed = time() - start_time
            hours = seconds_elapsed // 3600
            minutes = (seconds_elapsed - hours * 3600) // 60
            seconds = seconds_elapsed - hours * 3600 - minutes * 60
            time_str = f"{epoch + 1}/{num_epochs} ({int(hours)}h {int(minutes)}m {int(seconds)}s)"
            print(
                f"{time_str}, train loss = {loss.numpy() / start_train_loss - 1:.4%}, validation_loss = {validation_loss.numpy() / start_val_loss - 1:.4%}, train + validation loss = {(loss.numpy() + validation_loss.numpy()) / (start_train_loss + start_val_loss) - 1:.4%}"
            )
            if epoch > 0 and epoch % 20 == 0:
                self.training_log = loss_calculator.training_log
                self.save()

        self.training_log = loss_calculator.training_log

    def get_weights(self, year: int):
        return np.maximum(
            0,
            self.sim.calc("household_weight", period=year).values
            + self.weight_changes[year - self.start_year],
        )

    def get_microsimulation(self):
        sim_reweighted = Microsimulation(adjust_weights=False)
        for year in range(self.start_year, self.end_year + 1):
            new_weights = np.maximum(
                0,
                self.sim.calc("household_weight", period=year).values
                + self.weight_changes[year - self.start_year],
            )
            sim_reweighted.set_input(
                "household_weight",
                year,
                new_weights,
            )
        return sim_reweighted

    def save(
        self, folder: Path = policyengine_uk_MICRODATA_FOLDER, run_id: str = 1
    ):
        if isinstance(folder, str):
            folder = Path(folder)

        for period in range(self.start_year, self.end_year + 1):
            CalibratedFRS.save(
                self.start_year,
                f"household_weight/{period}",
                self.get_weights(period),
            )

        log = pd.DataFrame(self.training_log)
        log["run_id"] = run_id
        log.to_csv(folder / f"training_log_run_{run_id}.csv", index=False)


if __name__ == "__main__":
    parser = ArgumentParser()
    # Arguments for epochs, validation split and learning rate
    parser.add_argument(
        "--epochs",
        type=int,
        default=500,
        help="Number of epochs to run",
    )
    parser.add_argument(
        "--validation-split",
        type=float,
        default=0.0,
        help="Percentage of metrics to use as validation",
    )
    parser.add_argument(
        "--learning-rate",
        type=float,
        default=1e3,
        help="Learning rate for optimiser",
    )
    parser.add_argument(
        "--k-fold-cross-validation",
        type=int,
        default=1,
        help="Number of folds for k-fold cross validation",
    )
    args = parser.parse_args()

    if args.k_fold_cross_validation > 1:
        sim = Microsimulation(average_parameters=True, adjust_weights=False)
        loss_calculators = LossCalculator.create_k_fold_cv_calculators(
            sim, k=args.k_fold_cross_validation
        )
        for i in range(args.k_fold_cross_validation):
            print(f"Running fold {i + 1}/{args.k_fold_cross_validation}")
            weights = HouseholdWeights()
            weights.calibrate(
                num_epochs=args.epochs,
                validation_split=args.validation_split,
                learning_rate=args.learning_rate,
                loss_calculator=loss_calculators[i],
            )
            weights.save(run_id=i + 1)
        print("Combining individual fold logs")
        logs = pd.concat(
            [
                pd.read_csv(
                    REPO / "calibration" / f"training_log_run_{i + 1}.csv"
                )
                for i in range(args.k_fold_cross_validation)
            ]
        )
        logs.sort_values(["run_id", "epoch", "category", "validation"]).to_csv(
            REPO / "calibration" / "training_log.csv", index=False
        )
        for filename in glob(
            (REPO / "calibration" / "training_log_run_*.csv").as_posix()
        ):
            os.remove(filename)
    else:
        weights = HouseholdWeights()
        weights.calibrate(
            num_epochs=args.epochs,
            validation_split=args.validation_split,
            learning_rate=args.learning_rate,
        )
        weights.save()
    print("Calibration completed.")



from policyengine_uk.model_api import *


class gross_financial_wealth(Variable):
    label = "Gross financial wealth"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP


class net_financial_wealth(Variable):
    label = "Net financial wealth"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP



from typing import Iterable

import numpy as np
from ...loss_category import LossCategory
from .country_level_program import (
    UniversalCredit,
    ChildBenefit,
    ChildTaxCredit,
    WorkingTaxCredit,
    PensionCredit,
    IncomeSupport,
    StatePension,
    HousingBenefit,
    ESAIncome,
    JSAIncome,
    CouncilTax,
    TotalNI,
    EmploymentIncome,
    SelfEmploymentIncome,
    PensionIncome,
    SavingsInterestIncome,
    PropertyIncome,
    DividendIncome,
)
from .income_tax import IncomeTax


class Programs(LossCategory):
    name = "Programs"
    category = "Programs"
    weight = 2

    def initialise(self):
        self.initial_value = None
        self.subcategories = []
        for subcategory in (
            IncomeTax,
            UniversalCredit,
            ChildBenefit,
            ChildTaxCredit,
            WorkingTaxCredit,
            PensionCredit,
            IncomeSupport,
            StatePension,
            HousingBenefit,
            ESAIncome,
            JSAIncome,
            CouncilTax,
            TotalNI,
            EmploymentIncome,
            SelfEmploymentIncome,
            PensionIncome,
            SavingsInterestIncome,
            PropertyIncome,
            DividendIncome,
        ):
            self.subcategories.append(
                subcategory(
                    self.years,
                    self.year,
                    np.log(
                        self.sim.calc(
                            subcategory.variable, period=self.year
                        ).sum()
                        + np.e
                    ),
                    self.sim,
                )
            )

    def compute(self, *args, **kwargs):
        losses = 0
        log = []
        for subcategory in self.subcategories:
            additional_loss, additional_log = subcategory.compute(
                *args, **kwargs
            )
            losses += additional_loss * self.weight
            log += additional_log
        if self.initial_value is None:
            self.initial_value = losses
        losses /= self.initial_value
        for entry in log:
            entry["loss"] /= self.initial_value
        return losses, log

    def get_metric_names(self) -> Iterable[str]:
        names = sum(
            [
                subcategory.get_metric_names()
                for subcategory in self.subcategories
            ],
            [],
        )
        dated_names = []
        for name in names:
            for year in self.years:
                dated_names.append(f"{name}.{year}")
        return dated_names



from policyengine_uk.model_api import *


class marginal_tax_rate(Variable):
    label = "Marginal tax rate"
    documentation = "Percent of marginal income gains that do not increase household net income."
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = "/1"

    def formula(person, period, parameters):
        mtr_values = np.zeros(person.count, dtype=np.float32)
        simulation = person.simulation
        adult_index_values = person("adult_index", period)
        DELTA = 1_000
        for adult_index in [1, 2]:
            alt_simulation = simulation.get_branch(
                f"adult_{adult_index}_pay_rise"
            )
            mask = adult_index_values == adult_index
            for variable in simulation.tax_benefit_system.variables:
                if variable not in simulation.input_variables:
                    alt_simulation.delete_arrays(variable)
            alt_simulation.set_input(
                "employment_income",
                period,
                person("employment_income", period) + mask * DELTA,
            )
            alt_person = alt_simulation.person
            household_net_income = person.household(
                "household_net_income", period
            )
            household_net_income_higher_earnings = alt_person.household(
                "household_net_income", period
            )
            increase = (
                household_net_income_higher_earnings - household_net_income
            )
            mtr_values += where(mask, 1 - increase / DELTA, 0)
        return mtr_values



from policyengine_uk.model_api import *


class earned_income(Variable):
    value_type = float
    entity = Person
    label = "Total earned income"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        COMPONENTS = [
            "employment_income",
            "self_employment_income",
            "pension_income",
        ]
        return add(person, period, COMPONENTS)


class market_income(Variable):
    value_type = float
    entity = Person
    label = "Market income"
    documentation = "Income from market sources"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        INCOME_VARIABLES = [
            "employment_income",
            "self_employment_income",
            "savings_interest_income",
            "dividend_income",
            "miscellaneous_income",
            "property_income",
            "pension_income",
            "private_transfer_income",
            "maintenance_income",
        ]
        income = add(person, period, INCOME_VARIABLES)
        return income - person("maintenance_expenses", period)


class gross_income(Variable):
    value_type = float
    entity = Person
    label = "Gross income, including benefits"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        COMPONENTS = [
            "market_income",
            "benefits",
        ]
        return add(person, period, COMPONENTS)


class household_gross_income(Variable):
    value_type = float
    entity = Household
    unit = GBP
    label = "Household gross income"
    definition_period = YEAR
    unit = GBP

    def formula(household, period, parameters):
        return add(
            household,
            period,
            ["household_market_income", "household_benefits"],
        )


class net_income(Variable):
    value_type = float
    entity = Person
    label = "Net income"
    documentation = "Market income, minus taxes, plus benefits"
    unit = GBP
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("gross_income", period) - person("tax", period)


class hours_worked(Variable):
    value_type = float
    entity = Person
    label = "Total amount of hours worked by this person"
    definition_period = YEAR
    unit = "hour"


class in_work(Variable):
    value_type = bool
    entity = Person
    label = "Worked some hours"
    definition_period = YEAR

    def formula(person, period, parameters):
        has_hours_worked = person("hours_worked", period) > 0
        earnings = add(
            person, period, ["employment_income", "self_employment_income"]
        )
        has_earnings = earnings > 0
        return has_hours_worked | has_earnings


class weekly_hours(Variable):
    value_type = float
    entity = Person
    label = "Weekly hours"
    documentation = "Average weekly hours worked"
    definition_period = YEAR
    unit = "hour"
    quantity_type = FLOW

    def formula(person, period, parameters):
        return person("hours_worked", period) / WEEKS_IN_YEAR


class EmploymentStatus(Enum):
    FT_EMPLOYED = "Full-time employed"
    PT_EMPLOYED = "Part-time employed"
    FT_SELF_EMPLOYED = "Full-time self-employed"
    PT_SELF_EMPLOYED = "Part-time self-employed"
    UNEMPLOYED = "Unemployed"
    RETIRED = "Retired"
    STUDENT = "Student"
    CARER = "Carer"
    LONG_TERM_DISABLED = "Long-term sick/disabled"
    SHORT_TERM_DISABLED = "Short-term sick/disabled"
    OTHER_INACTIVE = "Inactive for another reason"
    CHILD = "Child"


class employment_status(Variable):
    value_type = Enum
    entity = Person
    possible_values = EmploymentStatus
    default_value = EmploymentStatus.UNEMPLOYED
    label = "Employment status of the person"
    definition_period = YEAR


class capital_income(Variable):
    value_type = float
    entity = Person
    label = "Income from savings or dividends"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        return add(
            person, period, ["savings_interest_income", "dividend_income"]
        )


class hbai_household_net_income(Variable):
    value_type = float
    entity = Household
    label = "Household net income (HBAI definition)"
    documentation = "Disposable income for the household, following the definition used for official poverty statistics"
    unit = GBP
    definition_period = YEAR

    def formula(household, period, parameters):
        gross_income = household("household_gross_income", period)
        tax = household("household_tax", period)
        excluded_income = household("baseline_hbai_excluded_income", period)
        return gross_income - tax - excluded_income


class household_net_income(Variable):
    label = "net income"
    documentation = "Household net income after taxes and benefits"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP
    adds = ["household_market_income", "household_benefits"]
    subtracts = ["household_tax"]


class real_household_net_income(Variable):
    label = "Real household net income"
    documentation = "Disposable income in January 2015 prices"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        def cpi(period):
            return parameters(period).calibration.uprating.CPI

        multiplier = cpi("2015-01-01") / cpi(period)
        return household("household_net_income", period) * multiplier


class hbai_household_net_income_ahc(Variable):
    value_type = float
    entity = Household
    label = "Household net income, after housing costs"
    definition_period = YEAR
    unit = GBP

    def formula(household, period, parameters):
        income = household("hbai_household_net_income", period)
        housing_costs = household("housing_costs", period)
        return income - housing_costs


class equiv_household_net_income(Variable):
    value_type = float
    entity = Household
    label = "Equivalised household net income"
    definition_period = YEAR
    unit = GBP

    def formula(household, period, parameters):
        income = household("household_net_income", period)
        equivalisation = household("household_equivalisation_bhc", period)
        return income / equivalisation


class equiv_hbai_household_net_income(Variable):
    value_type = float
    entity = Household
    label = "Equivalised household net income (HBAI)"
    definition_period = YEAR
    unit = GBP

    def formula(household, period, parameters):
        income = household("hbai_household_net_income", period)
        equivalisation = household("household_equivalisation_bhc", period)
        return income / equivalisation


class equiv_hbai_household_net_income_ahc(Variable):
    value_type = float
    entity = Household
    label = "Equivalised household net income, after housing costs (HBAI)"
    definition_period = YEAR
    unit = GBP

    def formula(household, period, parameters):
        income = household("hbai_household_net_income_ahc", period)
        equivalisation = household("household_equivalisation_ahc", period)
        return income / equivalisation


class base_net_income(Variable):
    value_type = float
    entity = Person
    label = "Existing net income for the person to use as a base in microsimulation"
    definition_period = YEAR
    unit = GBP


class is_apprentice(Variable):
    value_type = bool
    entity = Person
    label = "In an apprenticeship programme"
    definition_period = YEAR
    default_value = False


class MinimumWageCategory(Enum):
    APPRENTICE = "Apprentice"
    UNDER_18 = "Under 18"
    BETWEEN_18_20 = "18 to 20"
    BETWEEN_21_22 = "21 to 22"
    BETWEEN_23_24 = "23 to 24"
    OVER_24 = "25 or over"


class minimum_wage_category(Variable):
    value_type = Enum
    possible_values = MinimumWageCategory
    default_value = MinimumWageCategory.OVER_24
    entity = Person
    label = "Minimum wage category"
    definition_period = YEAR

    def formula(person, period, parameters):
        age = person("age", period)
        return select(
            [
                person("is_apprentice", period),
                age < 18,
                (age >= 18) & (age <= 20),
                (age >= 21) & (age <= 22),
                (age >= 23) & (age <= 24),
                age >= 25,
            ],
            [
                MinimumWageCategory.APPRENTICE,
                MinimumWageCategory.UNDER_18,
                MinimumWageCategory.BETWEEN_18_20,
                MinimumWageCategory.BETWEEN_21_22,
                MinimumWageCategory.BETWEEN_23_24,
                MinimumWageCategory.OVER_24,
            ],
        )


class minimum_wage(Variable):
    value_type = float
    entity = Person
    label = "Minimum wage"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        MW = parameters(period).gov.hmrc.minimum_wage
        return MW[person("minimum_wage_category", period)]


class household_market_income(Variable):
    value_type = float
    entity = Household
    label = "market income"
    documentation = "Market income for the household"
    definition_period = YEAR
    unit = GBP
    adds = [
        "employment_income",
        "self_employment_income",
        "savings_interest_income",
        "dividend_income",
        "miscellaneous_income",
        "property_income",
        "pension_income",
        "private_transfer_income",
        "maintenance_income",
    ]


class household_income_decile(Variable):
    label = "household income decile"
    documentation = "Decile of household income (person-weighted)"
    entity = Household
    definition_period = YEAR
    value_type = int

    def formula(household, period, parameters):
        income = household("household_net_income", period)
        count_people = household("household_count_people", period)
        household_weight = household("household_weight", period)
        weighted_income = MicroSeries(
            income, weights=household_weight * count_people
        )
        return weighted_income.decile_rank().values


class income_decile(Variable):
    label = "income decile"
    documentation = "Decile of household net income. Households are sorted by disposable income, and then divided into 10 equally-populated groups."
    entity = Person
    definition_period = YEAR
    value_type = int

    def formula(person, period, parameters):
        return person.household("household_income_decile", period)



from policyengine_uk.model_api import *

label = "Housing"


class rent(Variable):
    label = "Rent"
    documentation = (
        "The total amount of rent paid by the household in the year."
    )
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP
    quantity_type = FLOW


class CouncilTaxBand(Enum):
    A = "A"
    B = "B"
    C = "C"
    D = "D"
    E = "E"
    F = "F"
    G = "G"
    H = "H"
    I = "I"


class council_tax_band(Variable):
    value_type = Enum
    possible_values = CouncilTaxBand
    default_value = CouncilTaxBand.D
    entity = Household
    label = "Council Tax band"
    definition_period = YEAR



import re
import shutil
from policyengine_core.data import PrivateDataset
import pandas as pd
from pathlib import Path
from tqdm import tqdm
from policyengine_uk.data.storage import policyengine_uk_MICRODATA_FOLDER


class RawLCFS(PrivateDataset):
    name = "raw_lcfs"
    label = "Raw LCFS"
    folder_path = policyengine_uk_MICRODATA_FOLDER
    is_openfisca_compatible = False

    filename_by_year = {
        2019: "raw_lcfs_2019.h5",
    }

    def generate(self, year: int, ukds_tab_zipfile: str):
        """Generates the raw LCFS tabular dataset from the TAB zip archive
        downloadable from the UKDS.

        Args:
            year (int): The year of the FRS to generate.
            ukds_tab_zipfile (str): The path to the TAB zip archive.
        """

        folder = Path(ukds_tab_zipfile)
        year = str(year)

        if not folder.exists():
            raise FileNotFoundError("Invalid path supplied.")

        new_folder = self.folder_path / "tmp"
        shutil.unpack_archive(folder, new_folder)
        folder = new_folder

        main_folder = next(folder.iterdir())
        tab_folder = main_folder / "tab"
        if tab_folder.exists():
            criterion = re.compile(".*\.tab")
            data_files = [
                path
                for path in tab_folder.iterdir()
                if criterion.match(path.name)
            ]
            task = tqdm(data_files, desc="Saving raw data tables")
            with pd.HDFStore(RawLCFS.file(year)) as file:
                for filepath in task:
                    task.set_description(
                        f"Saving raw data tables ({filepath.name})"
                    )
                    table_name = filepath.name.replace(".tab", "")
                    df = pd.read_csv(
                        filepath, delimiter="\t", low_memory=False
                    ).apply(pd.to_numeric, errors="coerce")
                    file[table_name] = df
        else:
            raise FileNotFoundError("Could not find the TAB files.")

        tmp_folder = self.folder_path / "tmp"
        if tmp_folder.exists():
            shutil.rmtree(tmp_folder)


RawLCFS = RawLCFS()



from policyengine_uk.model_api import *


class basic_income(Variable):
    label = "Basic income"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(person, period, parameters):
        bi_maximum = person("bi_maximum", period)
        bi_phaseout = person("bi_phaseout", period)
        return bi_maximum - bi_phaseout



from re import I
from policyengine_uk.situation_examples.examples import single_person_UC
from typing import List
from policyengine_uk.api import *
from policyengine_uk.graphs.data import get_wide_reform_individual_data
from policyengine_uk.graphs.general import (
    formalise_column_names,
    COLORS,
)
import plotly.express as px


def budget_chart(
    reforms: List[Reform],
    names: List[str] = None,
    variables: List[str] = ["household_net_income"],
    situation_function: Callable = single_person_UC,
    return_df: bool = False,
    **kwargs
):
    """Generates a plot of the budget changes for a list of reforms.

    Args:
        reforms (List[Reform]): A list of reform classes.
        names (List[str], optional): The reform names, if multiple reforms are provided.
        variables (List[str], optional): The variables to include in the budget lines. Defaults to ["household_net_income"].
        situation_function (Callable, optional): A function adding situational data to an IndividualSim. Defaults to single_person_UC.
        return_df (bool, optional): Whether to return the data used for the plots. Defaults to False.
        kwargs: Keyword arguments passed to fig.update_layout.

    Returns:
        Figure: The Plotly figure.
    """

    if not isinstance(reforms, list):
        reforms = [reforms]
        names = ["Reform"]

    df = get_wide_reform_individual_data(
        reforms,
        names,
        variables,
        situation_function,
        include_derivatives=False,
    )
    df, column_map = formalise_column_names(df)
    fig = px.line(
        df,
        x=column_map["employment_income"],
        y=[
            column_map[x]
            for x in column_map
            if ("baseline" in x or "reform" in x) and "deriv" not in x
        ],
        animation_frame="Reform",
        color_discrete_sequence=COLORS,
    )
    fig.update_layout(
        legend_title="Source",
    )
    fig.update_layout(**kwargs)
    if return_df:
        return fig, df
    return fig


INVERTED_DERIV_VARIABLES = (
    "household_net_income",
    "net_income",
    "benefits",
    "universal_credit",
)


def mtr_chart(
    reforms: List[Reform],
    names: List[str] = None,
    variables: List[str] = ["household_net_income"],
    situation_function: Callable = single_person_UC,
    invert: List[str] = [],
    derivatives: bool = False,
    return_df: bool = False,
    **kwargs
):
    """Generates a plot of the marginal tax rates for given reforms on a scenario.

    Args:
        reforms (Union[Reform, List[Reform]]): The reforms (multiple reforms creates a slider).
        names (List[str], optional): The reform names, if multiple reforms are provided.
        variables (List[str], optional): The variables to include MTRs or derivatives for, useful for explainable outputs. Defaults to ["household_net_income"].
        situation_function (Callable, optional): A function adding situational data to an IndividualSim. Defaults to single_person_UC.
        invert (List[str], optional): A list of variables that should be inverted to become a 'tax' from a derivative - e.g. NI should not be inverted, but benefits should. Defaults to [].
        derivatives (bool, optional): Whether to plot derivatives or MTRs. The difference is that in an MTR plot, higher values mean higher net incomes, but this isn't true for a derivative plot. Defaults to False.
        return_df (bool, optional): Whether the return the data used to generate the plot. Defaults to False.
        kwargs: Keyword arguments passed to fig.update_layout.

    Returns:
        Figure: The Plotly figure.
    """

    if not isinstance(reforms, list):
        reforms = [reforms]
        names = ["Reform"]

    test_sim = IndividualSim()
    situation_function(test_sim)
    situation_data = test_sim.situation_data
    households = situation_data["households"]
    adults = households[list(households.keys())[0]]["adults"]
    primary_adult = adults[0]

    invert += INVERTED_DERIV_VARIABLES
    df = get_wide_reform_individual_data(
        reforms,
        names,
        variables,
        situation_function,
        primary_adult_name=primary_adult,
        include_derivatives=True,
    )
    for var in df.columns:
        if not derivatives and (
            var.replace("baseline_", "")
            .replace("reform_", "")
            .replace("_deriv", "")
            in invert
        ):
            df[var] = -df[var]
            if df[var].min() < 0 and df[var].max() <= 0:
                df[var] += 1

    df.drop(
        [
            col
            for col in df.columns
            if (
                (("baseline_" in col) or ("reform_" in col))
                and ("deriv" not in col)
            )
        ],
        axis=1,
        inplace=True,
    )

    df, column_map = formalise_column_names(df)
    fig = px.line(
        df,
        x=column_map["employment_income"],
        y=[
            column_map[x]
            for x in column_map
            if ("baseline" in x or "reform" in x) and "deriv" in x
        ],
        animation_frame="Reform",
        color_discrete_sequence=COLORS,
    )
    fig.update_layout(legend_title="Source", yaxis_tickformat="%")
    fig.update_layout(**kwargs)
    if return_df:
        return fig, df
    return fig



import logging
from policyengine_core.variables import Variable
from typing import Callable, Type
import h5py
from policyengine_uk.data.datasets.frs.enhanced.stages.imputation.enhanced_frs import (
    EnhancedFRS,
)
from policyengine_uk.repo import REPO


def baseline_is_nonzero(variable: Type[Variable]) -> Callable:
    """Creates an OpenFisca formula calculating the whether the specified
    variable is non-zero in the baseline.

    Args:
        variable (Type[Variable]): The variable to calculate the change in.

    Returns:
        Callable: The OpenFisca formula.
    """

    def formula(entity, period):
        return entity("baseline_" + variable.__name__, period) > 0

    formula.__doc__ = f"Baseline-requiring formula for {variable.__name__}"
    return formula


def change_over_baseline(variable: Type[Variable]) -> Callable:
    """Creates an OpenFisca formula calculating the change in the
    specified variable over its baseline value.

    Args:
        variable (Type[Variable]): The variable to calculate the change in.

    Returns:
        Callable: The OpenFisca formula.
    """

    def formula(entity, period):
        baseline_value = entity("baseline_" + variable.__name__, period)
        return entity(variable.__name__, period) - baseline_value

    formula.__doc__ = f"Baseline-requiring formula for {variable.__name__}"
    return formula


def generate_baseline_variables():
    """
    Save baseline values of variables to a H5 dataset.
    """

    from policyengine_uk import Microsimulation

    YEARS = list(range(2022, 2026))
    baseline = Microsimulation(add_baseline_values=False)

    variable_metadata = baseline.simulation.tax_benefit_system.variables

    for variable in variable_metadata:
        if variable[:9] == "baseline_":
            for year in YEARS:
                baseline.simulation.set_input(
                    variable,
                    year,
                    [True]
                    * len(
                        baseline.calc(
                            variable_metadata[variable].entity.key + "_id"
                        )
                    ),
                )

    # First, find variables which need baseline storage.

    variables = list(
        filter(
            lambda variable: hasattr(variable, "formula")
            and variable.formula.__doc__ is not None
            and "Baseline-requiring formula" in variable.formula.__doc__,
            variable_metadata.values(),
        )
    )
    variables = list(
        map(
            lambda variable: variable_metadata[
                variable.formula.__doc__.split(" for ")[1]
            ],
            variables,
        )
    )
    print(f"Found {len(variables)} variables to store baseline values for:")
    print("\n* " + "\n* ".join([variable.label for variable in variables]))

    with h5py.File(EnhancedFRS.file(2022), "w") as f:
        for year in YEARS:
            for variable in variables:
                f.create_dataset(
                    f"baseline_{variable.name}/{year}",
                    data=baseline.calc(variable.name, period=year),
                )


if __name__ == "__main__":
    generate_baseline_variables()



# -*- coding: utf-8 -*-

import os
from policyengine_uk import entities
from policyengine_uk.system import (
    CountryTaxBenefitSystem,
    Microsimulation,
    Simulation,
    IndividualSim,
    COUNTRY_DIR,
    BASELINE_VARIABLES,
    parameters,
    variables,
)
from pathlib import Path
import os
from policyengine_core.taxbenefitsystems import TaxBenefitSystem

from policyengine_uk.data import (
    FRS,
    RawFRS,
    SPI,
    RawSPI,
    RawLCFS,
    RawWAS,
    ExtendedFRS,
    LCFS,
    WAS,
    CalibratedFRS,
    EnhancedFRS,
    SynthFRS,
)

REPO = Path(__file__).parent

DATASETS = [EnhancedFRS, SynthFRS, FRS, RawFRS]



from policyengine_uk import Microsimulation
from policyengine_uk.data import EnhancedFRS
import pytest
from itertools import product

UPRATED_VARIABLES = (
    "council_tax",
    "employment_income",
    "pension_income",
    "self_employment_income",
    "property_income",
    "savings_interest_income",
    "dividend_income",
)

sim = Microsimulation(dataset=EnhancedFRS, dataset_year=2022)


@pytest.mark.parametrize(
    "variable,year", product(UPRATED_VARIABLES, range(2023, 2026))
)
def test_uprating(variable: str, year: int):
    assert any(
        sim.calc(variable, period=year).values
        != sim.calc(variable, period=2022).values
    )



from policyengine_uk.entities import *
from policyengine_uk.tools.general import *
from policyengine_core.model_api import *
from policyengine_core import periods
from policyengine_uk import (
    reforms,
    BASELINE_VARIABLES,
    Microsimulation,
    IndividualSim,
)
import policyengine_uk.graphs as graphs



from policyengine_core.parameters import ParameterNode, Parameter
from policyengine_core.periods import instant


def average_parameters(parameters: ParameterNode) -> ParameterNode:
    """Averages parameter values over each year."""
    for parameter in parameters.get_descendants():
        if isinstance(parameter, Parameter):
            for year in range(2019, 2029):
                values = []
                for month in range(1, 13):
                    values.append(parameter(instant(f"{year}-{month:02}-01")))
                try:
                    parameter.update(
                        period=f"year:{year}:1", value=sum(values) / 12
                    )
                except:
                    pass

    return parameters



from policyengine_core.parameters import ParameterNode, Parameter


def add_price_cap_parameters(
    parameters: ParameterNode,
    start_year: 2022,
    start_quarter: 4,
    end_year: 2023,
    end_quarter: 4,
) -> ParameterNode:
    """Add price cap parameters to the parameters tree.

    Args:
        parameters (ParameterNode): The parameters tree.
        start_year (int): The start year of the price cap.
        start_quarter (int): The start quarter of the price cap.
        end_year (int): The end year of the price cap.
        end_quarter (int): The end quarter of the price cap.

    Returns:
        ParameterNode: The parameters tree with price cap parameters added.
    """

    price_cap: ParameterNode = parameters.gov.ofgem.price_cap

    for year in range(start_year, end_year + 1):
        for quarter in range(1, 5):
            if year == start_year and quarter < start_quarter:
                continue
            if year == end_year and quarter > end_quarter:
                break
            name = f"{year}_q{quarter}"
            value = price_cap.base(f"{year}-{quarter * 3 - 2:02}-01")
            new_parameter = Parameter(
                name=name,
                data=dict(
                    description=f"Default tariff price cap for the average household in {year} Q{quarter}.",
                    values={
                        "0000-01-01": value,
                    },
                    metadata=dict(
                        name=f"ofgem_price_cap_{name}",
                        label=f"Ofgem energy price cap ({year} Q{quarter})",
                        unit="currency-GBP",
                        period="year",
                    ),
                ),
            )
            new_parameter.name = "gov.ofgem.price_cap." + new_parameter.name
            price_cap.add_child(
                name,
                new_parameter,
            )

    return parameters



from policyengine_uk.model_api import *


class cost_of_living_support_payment(Variable):
    label = "Cost-of-living support payment"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        col = parameters(period).gov.treasury.cost_of_living_support
        on_means_tested_benefits = (
            add(
                household,
                period,
                col.means_tested_households.qualifying_benefits,
            )
            > 0
        )
        means_test_bonus = (
            col.means_tested_households.amount * on_means_tested_benefits
        )

        on_pensioner_benefits = (
            add(household, period, col.pensioners.qualifying_benefits) > 0
        )
        pensioner_bonus = col.pensioners.amount * on_pensioner_benefits

        on_disability_benefits = (
            add(household, period, col.disabled.qualifying_benefits) > 0
        )
        disabled_bonus = col.disabled.amount * on_disability_benefits

        return means_test_bonus + pensioner_bonus + disabled_bonus
